from numpy import abs
from numpy import array
from numpy import arange
from numpy import zeros

from scipy.sandbox.svm import RBFKernel
from scipy.sandbox.svm import LibSvmNuRegressionModel
from scipy.sandbox.svm import LibSvmTestDataSet

from gumby.datagenerator.temporal import OneDTemporalGenerator
from gumby.datagenerator.temporal import make_harmonic

from pylab import (
    figure, subplot, title, grid, xlabel, ylabel, plot, concatenate, show)


def predict_signal(svm, past, future_size=10):
    """Use a trained SVM and an initial signal to predict its evolution"""
    past = list(past)
    future = zeros(future_size, 'd')

    for i in xrange(future_size):
        next_prediction = svm.predict(LibSvmTestDataSet(array([past])))
        future[i] = next_prediction
        past.pop(0)
        past.append(next_prediction)

    return array(future)


def compute_error(signal1, signal2, window_size=10):
    """Compute the absolute difference, averaged on a sliding window"""
    abs_error = abs(signal1 - signal2)
    if window_size in (1, None):
        return abs_error
    abs_error = concatenate((zeros(window_size, abs_error.dtype), abs_error))

    averaged_error = []
    for i in xrange(len(signal1)):
        averaged_error.append(
            abs_error[i:i + window_size].sum() / window_size)
    return averaged_error


if __name__ == '__main__':
    model_type = LibSvmNuRegressionModel
    c = 100

    # parameters are by order:
    #    - amplitude
    #    - period in time steps
    #    - offset in time steps
    #    - mean value

    params = [
        [
            (1, 50, 0, -0.3)
        ],
        [
            (3, 100, 0, 0.2),
            (0.5, 50, 20),
        ],
        [
            (0.5, 203, 0),
            (0.5, 200, 20),
            (0.5, 25, 5),
        ],
        [
            (0.5, 50, 0, 1),
            (0.5, 20, 12),
            (0.2, 30, 56),
            (0.2, 35, 56),
            (0.2, 36, 56),
            (0.2, 37, 52),
            #(0.2, 47, 52),
        ],
    ]
    funcs = [make_harmonic(param) for param in params]

    # training the model on 3 past signals
    past_size = 500
    memory_size = 100

    generator = OneDTemporalGenerator(
        funcs, past_size, memory_size, randomized=True, differential=False)

    _, _, traindata = generator.make_training_set()

    model = model_type(RBFKernel(traindata.gamma), cost=c)
    svm = model.fit(traindata)

    # predicting the future of the first signal
    future_size = 1000
    for i, f in enumerate(funcs):

        future_seed, _, _ = generator.make_training_set([f], size=1,
                                                        t0=past_size-1)
        predicted_future = predict_signal(svm, future_seed[0], future_size)

        future_time = arange(future_size) + past_size + memory_size
        expected_future = f(future_time)

        figure(i)
        subplot(211)
        grid(True)
        title("Harmonic signal: %s" % params[i])
        xlabel('time')
        ylabel('signal value')
        plot(expected_future, 'r-')
        plot(predicted_future, 'b-')

        window_size = memory_size * 3
        averaged_error = compute_error(predicted_future, expected_future,
                                       window_size)
        subplot(212)
        xlabel('time')
        ylabel('averaged error')
        grid(True)
        plot(averaged_error, 'b-')

    show()


