# (C) Copyright 2007 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""Use a sequence of PIL Image instances as Brain input"""

from PIL import Image
from itertools import cycle
from itertools import izip
from numpy import array
from numpy import cumsum
import cPickle as pickle
import os
import shutil
import tempfile

from gumby.brain import Brain
from gumby.plotting import make_plot
from gumby.plotting import save_plot


def get_image_dimensions(img):
    """Utility function to compute the brain dimension to use given the mode"""
    if len(img.getbands()) == 3:
        return img.size + (3,)
    else:
        return img.size


class Movie(object):
    """A sequence of images with semantic description"""

    avi_cmd_pattern = ("mencoder 'mf://%s/*.png' -mf type=png:fps=%0.2f "
                       "-ovc lavc -lavcopts vcodec=mpeg4 -oac copy -o %s")

    gif_cmd_pattern = "convert -delay %d %s/*.png %s"

    avi_play_cmd_pattern = "vlc %s"

    def __init__(self, images, fps, descriptions=(),  name="movie"):
        # sequence of a PIL image objects
        self._images = images

        # sequence of semantic descriptions of the previous images
        self._descriptions = descriptions

        self.fps = fps
        self.name = name

    #
    # pickling support
    #

    def __getstate__(self):
        """Override gestate to make the movie picklable"""
        dict_ = self.__dict__.copy()


        dict_['_images'] = [(img.mode, img.size, img.tostring())
                            for img in dict_['_images']]
        return dict_

    def __setstate__(self, dict_):
        """Override setstate to load pickled movies"""
        dict_['_images'] = [Image.fromstring(m, s, d)
                            for m, s, d in dict_['_images']]
        self.__dict__.update(dict_)

        # compat with old definition
        if not hasattr(self, 'fps'):
            self.fps = self._fps

    def save(self, filename=None, directory=None):
        """Save the complete movie instance as a pickle"""
        filename = filename or self.name + '.pickle'
        if directory:
            filename = os.path.join(directory, filename)
        pickle.dump(self, file(filename, 'w'), -1)

    #
    # sequence / iterator interface
    #

    def __len__(self):
        return len(self._images)

    def __iter__(self):
        return izip(self._images, self._descriptions)

    def __getitem__(self, index):
        return self._images[index]

    #
    # Brain interoperability
    #

    def get_dimensions(self):
        """Return the 2D dimensions of the first frame + 3 if in RGB"""
        return get_image_dimensions(self._images[0])

    #
    # Image / movies export / preview
    #

    def save_as_img_sequence(self, directory, format='png'):
        filenames = []
        for i, img in enumerate(self._images):
            filename = os.path.join(directory, self.name + '_%04d.%s' % (
                i, format))
            img.save(filename)
            filenames.append(filename)
        return filenames

    def save_as_avi_movie(self, filename=""):
        filename = filename or self.name + '.avi'
        directory = tempfile.mkdtemp()
        self.save_as_img_sequence(directory)
        cmd = self.avi_cmd_pattern % (directory, self.fps, filename)
        os.system(cmd)
        shutil.rmtree(directory)

    def save_as_animated_gif(self, filename):
        filename = filename or self.name + '.gif'
        directory = tempfile.mkdtemp()
        self.save_as_img_sequence(directory)
        os.system(self.gif_cmd_pattern % (
            1 / self.fps * 100, directory, filename))
        shutil.rmtree(directory)

    #
    # Basic playback for inline debugging
    #

    def play(self):
        _, filename = tempfile.mkstemp()
        self.save_as_avi_movie(filename)
        cmd = self.avi_play_cmd_pattern % filename
        os.system(cmd)
        os.remove(filename)


class StereoMovie(object):
    """Use two cameras to shoot two streams of images for the same scene"""
    # TODO
    pass


def load_movie(filename):
    """Load a saved movie pickle"""
    return pickle.load(file(filename, 'r'))


def movavg(s, n):
    """Return an n period moving average for the time series s

    s is a list ordered from oldest (index 0) to most recent (index -1)
    n is an integer
    """
    s = array(s)
    c = cumsum(s)
    return (c[n-1:] - c[:-n+1]) / float(n)


class MovieAdapter(object):
    """Adapter to make a brain able to watch Coca-Cola(tm) ads on TV

    This implements the ``next`` method of the python generator interface to
    inject and predict the next movie frame.
    """

    record = True

    def __init__(self, movie, brain=None, named_nodes=None, dimensions=None,
                 record=True):

        # TODO: make it possible to adapt the movie size to the brain dimensions
        self._movie = movie
        self.dimensions = dimensions or movie.get_dimensions()

        # TODO: replace Brain by any IPredictor
        if brain is None:
            # build a single layer brain with dimensions mathcing the movie's
            self.brain = Brain()
            self.brain.add_layer(dimensions=self.dimensions)
        else:
            # check input dimensions compatibility here
            self.brain = brain

        self.mode = len(self.dimensions) == 3 and 'RGB' or 'L'
        self.record = record
        self._recorded_frames = dict()
        self._recorded_mse = dict()

        # initialize the frame generator and ensure the brain has no previous
        # historical data
        self.reset()

        return brain

    def img_to_array(self, image):
        """Convert a PIL Image into a numpy array suitable as brain input"""
        if image.mode != self.mode:
            image = image.convert(mode=self.mode)
        if image.size != self.dimensions[:2]:
            image = image.resize(self.dimensions[:2], Image.ANTIALIAS)

        w, h = self.dimensions[:2]
        data = array(image.getdata())
        rescaled_data = data / 128. - 1
        return rescaled_data.reshape(h, w).transpose()

    def array_to_img(self, data, reference_image):
        """Convert brain ouput to an Image with the same format as the input"""

        rescaled = ((data + 1.) * 128).clip(0, 255)

        image = Image.new(self.mode, (rescaled.shape[0], rescaled.shape[1]))
        image.putdata(list(rescaled.transpose().flat))
        if image.size != reference_image.size:
            image = image.resize(reference_image.size, Image.NEAREST)
        return image

    def build_frame_generator(self):
        """Python generator to iteratively transcode img to brain I/O"""
        for frame, description in self._movie:
            # XXX: handle named_nodes
            input = self.img_to_array(frame)
            output = self.brain.predict([input])
            if self.record:
                self.record_current_state()
            # TODO: the [0] assumes we only use a single layer
            yield self.array_to_img(output[0], frame)

    def record_current_state(self):
        """"Add the current state of to the record history"""
        reference, _ = iter(self._movie).next()
        for l in self.brain:
            recorded_frames = self._recorded_frames.setdefault(l, [])
            recorded_frames.append(
                self.array_to_img(l.get_prediction(), reference))
            recorded_mse = self._recorded_mse.setdefault(l, [])
            recorded_mse.append(l.get_squared_prediction_error())

    def get_movie(self, layer):
        """Return a movie instance of the selected layer"""
        if isinstance(layer, int):
            layer_id = layer
            layer = list(self.brain.layers)[layer]
        else:
            layer_id = id(layer)

        recorded_frames = self._recorded_frames.get(layer)
        if recorded_frames is None:
            return None
        return Movie(recorded_frames, self._movie.fps,
                     name="Recorded frames for layer %d" % layer_id)

    def get_movies(self):
        """Return the movies recorded for each layer of the brain"""
        return [self.get_movie(l) for l in self.brain.get_ranked_layers()]

    def get_mse(self, layer=None):
        """Return the recorded historical mean squared errors for layer

        If no specific layer is provided return the list of all recorded
        MSEs in ranked layers order.
        """
        if layer is not None:
            return self._recorded_mse.get(layer)
        else:
            return [self._recorded_mse.get(l)
                    for l in self.brain.get_ranked_layers()]

    def plot_mse(self, filename, movavg_period=0):
        """Plot the recorded mse into files


        If movavg_period > 0, the plotted data is averaged over a window
        with the requested period.
        """
        plot = make_plot("Recorded Mean Square Errors")
        for i, layer in enumerate(self.brain.get_ranked_layers()):
            mse = self._recorded_mse.get(layer)
            if mse is not None:
                label = "Layer %d" % i
                if movavg_period > 0:
                    plot.semilogy(movavg(mse, movavg_period), label=label)
                else:
                    plot.semilogy(mse, label=label)
        plot.legend()
        save_plot(plot, filename)

    def next(self):
        """Watch the next frame and predict the outcome as an Image instance"""
        return self._frame_generator.next()

    def reset(self):
        self._recorded_frames.clear()
        self._recorded_mse.clear()
        self.brain.reset()
        self._frame_generator = self.build_frame_generator()

    def train(self, nb_passes=None, find_parameters=False,
              max_sample_size=1000, **kw):
        """Train the brain using the movie as training set

        TODO: explain the training scheme here
        """

        if nb_passes is None:
            # default one pass by layer in the brain
            nb_passes = len(self.brain)

        self.reset()
        dataset = cycle(self.img_to_array(f) for f, d in self._movie)
        for pass_index in xrange(nb_passes):
            for layer in self.brain.get_ranked_layers():
                # load the layer historical data with enough data from the
                # movie
                for i in xrange(layer.history_size):
                    self.brain.predict([dataset.next()])
                    if self.record:
                        self.record_current_state()

                layer.train(find_parameters=find_parameters,
                            max_sample_size=max_sample_size, **kw)


