# (C) Copyright 2007-2008 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""The gumby brain: layered set of gumby nodes"""

from gumby.history import History
from gumby.model import SvmModel
from gumby.odict import OrderedDict
from itertools import izip
from itertools import cycle
import logging
from numpy import array
from numpy import ndindex
from numpy import zeros
import random
from time import time


def sample_max(inputs, outputs, max_sample_size):
    sampled = random.sample(zip(inputs, outputs), max_sample_size)
    return [i for i, _ in sampled], [o for _, o in sampled]


class Layer(object):
    """An array of gumby cells sharing the same predictive model"""

    logger = logging.getLogger("gumby.brain.Layer")

    # TODO: make the following properties to update the model size uppon
    # change

    # spatial neighborhood sizes accross all dimensions used for prediction
    # should be a tuple of odd integers (to allow for spatial centering)
    spatial_neighborhood = None

    # temporal neighborhood size used for prediction
    temporal_neighborhood = 3

    model = None

    # TODO: replace by a new connector adapter class that can handle all sorts
    # of mapping logics
    _target_layers = ()

    def __init__(self, dimensions=(10,), history_size=10,
                 spatial_neighborhood=None, temporal_neighborhood=3):
        self.dimensions = dimensions
        if spatial_neighborhood is None:
            self.spatial_neighborhood = (3,) * len(dimensions)
        else:
            assert len(spatial_neighborhood) == len(dimensions)
            self.spatial_neighborhood = spatial_neighborhood

        assert temporal_neighborhood > 0
        self.temporal_neighborhood = temporal_neighborhood

        self._update_model()

        self.history_size = history_size

        self._surprise_history = History(
            dimensions, self.spatial_neighborhood, history_size)
        self._prediction_history = History(
            dimensions, self.spatial_neighborhood, history_size)
        self._input_history = History(
            dimensions, self.spatial_neighborhood, history_size)

    def reset(self):
        """Set all historical data"""
        self._surprise_history.reset()
        self._prediction_history.reset()
        self._input_history.reset()

    def _update_model(self):
        layers = (self,) + self._target_layers
        model_input_size = 0
        for layer in layers:
            l_model_input_size = reduce(lambda x, y: x * y,
                                        layer.spatial_neighborhood)
            if layer is self:
                l_model_input_size *= layer.temporal_neighborhood
            model_input_size += l_model_input_size

        assert model_input_size != 0, "model input_size cannot be 0"

        if self.model is None or self.model.input_size != model_input_size:
            # TODO: short term: reuse old model characteristics
            # TODO: medium term: make models able to increase
            # there input size without loosing existing support
            # vectors and parameters by padding the SVs with zeros
            self.model = SvmModel(model_input_size)

    def __repr__(self):
        return "Layer(dimensions=%r, history_size=%r)" % (
            self.dimensions, self.history_size)

    def __len__(self):
        number_of_cells = reduce(lambda x, y: x * y,
                                  self.dimensions, 1)
        return number_of_cells * self.history_size

    def add_target_layer(self, target_layer):
        self._target_layers += (target_layer,)
        self._update_model()

    def get_surprise(self, time_index=0):
        """Return the stored surprise array of the given time index"""
        return self._surprise_history[time_index]

    def get_prediction(self, time_index=0):
        """Return the stored prediction array of the given time index"""
        return self._prediction_history[time_index]

    def get_input(self, time_index):
        """Return the stored input array of the given time index"""
        return self._input_history[time_index]

    def predict(self, inputs):
        """Compute the next prediction array

        Each elements is the result of the SVM prediction of the layer model fed
        with the values of the spatial and temporal neighborhood of the element.
        """
        t0 = time()
        self._input_history.record(inputs)

        # compute the new value of the surprise and store it in the first
        # position of the history
        new_surprise = self._prediction_history[0] - inputs
        self._surprise_history.record(new_surprise)

        # create a new array to hold the future predicted values
        new_prediction = zeros(self.dimensions)

        # build the list of neighborhood inputs formatted for the model for
        # every cell of the layer
        model_inputs = []
        for position in ndindex(self.dimensions):
            model_inputs.append(self._build_model_input(position))

        # call the model with the list of inputs and place the results in
        # their respective position in the prediction array of the layer
        single_predictions = self.model.predict(model_inputs)
        if self.dimensions == (1,):
            single_predictions = single_predictions.reshape(self.dimensions)
        for pos, pred in izip(ndindex(self.dimensions), single_predictions):
            new_prediction[pos] = pred

        # record the new prediction in the layer history and return it
        # to the caller
        self._prediction_history.record(new_prediction)
        self.logger.info("prediction on layer %d [%0.3fs]", id(self), time()-t0)
        return new_prediction

    def get_squared_prediction_error(self, t=0):
        """Normalized squared prediction error of the layer"""
        surprise = self._surprise_history[t]
        return (surprise ** 2).mean()

    def get_mse(self):
        """Averaged squarred prediction error on all historical data"""
        return (array(list(self._surprise_history)) ** 2).mean()

    def train(self, find_parameters=False, max_sample_size=1000, **kw):
        """Train the embedded model against the layer historical data

        ``find_parameters``:
            if True, use cross validation to perform parameter
            estimation

        ``nr_folds``:
            number of folds to use in cross validations performed during
            parameters estimation
        """
        t0 = time()
        # collect data to build the training set
        # TODO: factor out in an external method?
        m_outputs = []
        m_inputs = []
        for t in xrange(self.history_size - self.temporal_neighborhood - 1):
            for position in ndindex(self.dimensions):
                m_outputs.append(self._input_history[t][position])
                m_inputs.append(self._build_model_input(
                    position, time_origin=t + 1))

        if max_sample_size is not None and len(m_inputs) > max_sample_size:
            self.logger.info("sampling %d items out of %d", max_sample_size,
                              len(m_inputs))
            m_inputs, m_outputs = sample_max(
                m_inputs, m_outputs, max_sample_size)

        # perform the actual model training
        if find_parameters:
            self.model.find_parameters(m_inputs, m_outputs, **kw)
        self.model.train(m_inputs, m_outputs)
        self.logger.info("training on layer %d on %d data points [%0.3fs]",
                         id(self), len(m_inputs), time()-t0)

    def _build_model_input(self, position, time_origin=0):
        """Helper method to build an input vector for the model"""
        model_input = zeros(self.model.input_size)
        spatial_offset = len(model_input) / (self.temporal_neighborhood +
                                             len(self._target_layers))

        # local spatial neighborhoods for each time index in the local
        # temporal neighborhood
        for t in xrange(self.temporal_neighborhood):
            n = self._input_history.get_neighborhood(position, time_origin+t)
            model_input[t*spatial_offset:(t+1)*spatial_offset] = n.flat

        # last spatial neighborhood for each target layer
        i0 = self.temporal_neighborhood
        for i, layer in enumerate(self._target_layers):
            n = layer._prediction_history.get_neighborhood(
                position, time_origin)
            model_input[(i0+i)*spatial_offset:(i0+i+1)*spatial_offset] = n.flat

        return model_input


class Brain(object):
    """The gumby brain: layered set of gumby cells"""

    train_max_sample_size = 1000

    logger = logging.getLogger("gumby.brain.Brain")

    def __init__(self):
        self.layers = OrderedDict()
        self._source_layer = OrderedDict()
        self._target_layers = OrderedDict()
        self._invalidate_rank_caches()

    def reset(self):
        """Set all historical data in layers back to zero"""
        for layer in self.layers:
            layer.reset()

    def _invalidate_rank_caches(self):
        self.input_layers = []
        self._cached_ranked_layers = None

    #
    # Methods to introspect the brain structure
    #

    def __len__(self):
        return len(self.layers)

    def __iter__(self):
        return self.layers.iterkeys()

    def __contains__(self, layer):
        return layer in self.layers

    def get_size(self):
        """Total number of cells in the brain"""
        return reduce(lambda x, y: x + y, (len(l) for l in self.layers))

    def get_source_layer(self, layer):
        return self._source_layer.get(layer)

    def get_target_layers(self, layer):
        return self._target_layers.get(layer) or []

    def get_input_layers(self):
        return [l for l in self if self.get_source_layer(l) is None]

    #
    # Methods to build the brain structure
    #

    def add_layer(self, layer=None, dimensions=(10,), history_size=10):
        """Add a new layer to the brain; creating if necessary"""

        assert layer not in self, "%r already in brain" % layer

        # invalidate the ranked nodes caches
        self._invalidate_rank_caches()

        # register the layer in the complete list of layers of the brain
        if layer is None:
            layer = Layer(dimensions, history_size)

        self.layers[layer] = 1
        self._source_layer[layer] = None
        self._target_layers[layer] = []
        return layer

    def connect(self, source_layer, target_layer):
        """Add a link between two layers"""
        self._source_layer[target_layer] = source_layer
        tl = self._target_layers.setdefault(source_layer, [])
        tl.append(target_layer)
        source_layer.add_target_layer(target_layer)

    #
    # Methods to perform inference on a trained brain
    #

    def get_ranked_layers(self):
        """Collect ranked layers according to their topology

        Source / bottom / input layers come first.
        """
        # TODO: order the layers according to connection topology
        return list(self.layers)

    def predict(self, inputs):
        """Set the input on the source layers and perform inference

        ``inputs`` should be a list of arrays matching the dimensions of the
        input layers.

        Each layer predict tries to predict the next state surprise of the
        layer below (or of the next brain input) while using the
        last predictions of the layers above.

        The prediction of the input layers are returned as an a array.
        """
        for layer in self.get_ranked_layers():
            t0 = time()
            source_layer = self._source_layer.get(layer)
            if source_layer is None:
                layer.predict(inputs.pop())
            else:
                layer.predict(source_layer.get_surprise(0))
        return self.get_prediction()

    def get_prediction(self, time_index=0):
        """Read the prediction on the input layers

        ``time_index == 0`` points to the last prediction
        """
        return [l.get_prediction(time_index) for l in self.get_input_layers()]

    def get_surprise(self, time_index=0):
        """Read the prediction on the input layers

        ``time_index == 0`` points to the last surprise
        """
        return [l.get_surprise(time_index) for l in self.get_input_layers()]

    def get_squared_prediction_error(self):
        """Return the sum of the squared prediction errors of the layers"""
        return array([l.get_squared_prediction_error() for l in self]).sum()

    #
    # Methods to train the models
    #

    def train(self, do_parameters=False):
        """Naive training method that iterates over the ranked layers"""
        # TODO: rewrite this method to be able to use a dataset as tracking
        # inputs to refill historical data after each layer training step
        for layer in self.get_ranked_layers():
            layer.train(do_parameters)

    def get_training_generator(self, do_parameters=False):
        """Naive training generator method that cycles over the ranked layers"""
        for layer in cycle(self.get_ranked_layers()):
            yield layer.train(self._target_layers.get(layer), do_parameters)


class MyBrainHurts(Exception):
    """Exception raised when the structure of the brain is fucked up"""
    pass


