# (C) Copyright 2007 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""soya based movie generator to build artificial training data"""

from gumby.adapter.movie import Movie
from gumby.adapter.movie import load_movie
from random import Random
import soya
import time
from PIL import Image

# make the data subfolder available to soya
import gumby.datagenerator
soya.path.append(gumby.datagenerator.__path__[0] + "/data")


class MovingBody(soya.Body):
    """Soya body that wanders along the z=0 plane"""

    initial_f_speed = .05
    f_speed_range = (-.05, .05)
    initial_r_speed = .0
    r_speed_range = (-5., 5.)
    fs_change_prob = .2
    rs_change_prob = .2

    def __init__(self, parent, model_name, rng=None, position=None):
        soya.Body.__init__(self, parent, soya.Model.get(model_name))
        self.speed = soya.Vector(self, .0, .0, self.initial_f_speed)
        self.forward_speed = self.initial_f_speed
        self.rotation_speed = self.initial_r_speed
        self._rng = rng or Random()
        self.model_name = model_name
        if position is not None:
            self.set_xyz(*position)

    def begin_round(self):
        soya.Body.begin_round(self)

        # change direction
        if self._rng.random() < self.rs_change_prob:
            self.rotation_speed = self._rng.uniform(*self.r_speed_range)

        # change direction
        if self._rng.random() < self.fs_change_prob:
            forward_speed = abs(
                self.forward_speed + self._rng.uniform(*self.f_speed_range))
            self.speed = soya.Vector(self, 0., 0., forward_speed)

    def advance_time(self, proportion):
        soya.Body.advance_time(self, proportion)
        self.rotate_y(proportion * self.rotation_speed)
        self.add_mul_vector(proportion, self.speed)

    def get_description(self, camera):
	"""return metadata wrt the camera position"""

        # assume that camera and self share the same parents
        assert self.parent is camera.parent
        p = self.position() - camera.vector_to(self)
        s = self.speed
        description = {
            'model_name': self.model_name,
            'position': (p.x, p.y, p.z),
            'speed': (s.x, s.y, s.z),
        }
        # TODO: add color, size
        return description


class MovieMaker(object):
    """Generate a random 3D animated scene with randomly moving bodies"""

    models = ('caterpillar', 'pyramid', 'cube')
    soya_dimensions = (320, 240)
    dimensions = (80, 60)
    length = 20
    time_step = 0.1
    camera_y_range = (-3, 10)
    camera_z_range = (5, 10)
    light_position = (10, 5, 5)

    def __init__(self, name_prefix='', seed=None, models=None, mode=None):
        self._rng = Random(seed)
        self.name_prefix = name_prefix
        if models is not None:
            self.models = models
        self.mode = mode
        self.bodies = list()
        soya.init(
            width=self.soya_dimensions[0],
            height=self.soya_dimensions[1],
        )

        # create the scene and fiat lux
        self.scene = soya.World()
        self.light = soya.Light(self.scene)
        self.camera = soya.Camera(self.scene)

        self.camera.y = self._rng.uniform(*self.camera_y_range)
        self.camera.z = self._rng.uniform(*self.camera_z_range)
        self.set_camera_direction(0, 0, 0)

        # TODO: add a background

    def set_camera_direction(self, *direction):
        p = soya.Point(self.scene, *direction)
        self.camera.look_at(p)

    def add_body(self, model_name, position=None):
        body = MovingBody(self.scene, model_name, self._rng, position=position)
        self.bodies.append(body)
        return body

    def add_random_body(self, position=None):
        return self.add_body(self._rng.choice(self.models), position=position)

    def seed(self, seed=None):
        """Seed the internal rng instance, use None for time-based seeding"""
        self._rng.seed(seed)

    def render_movie(self, name=None):
        """Build a movie of a randomly moving body"""

        # light the scene
        self.light.set_xyz(*self.light_position)

        # set the brain's eye position :)
        soya.set_root_widget(self.camera)

        images = []
        descriptions = []

        main_loop = soya.MainLoop(self.scene)
        for i in range(self.length):
            time.sleep(self.time_step)
            main_loop.update()
            soya.render()

            # save the frame data using PIL
            img = soya.screenshot().resize(self.dimensions, Image.ANTIALIAS)
            if self.mode is not None and self.mode != img.mode:
                img = img.convert(mode=self.mode)
            images.append(img)

            # save the descriptions metadata
            descriptions.append([body.get_description(self.camera)
                                 for body in self.bodies])

        if name == None:
            name = self.name_prefix
            if self.bodies:
                name += self.bodies[0].model_name
        return Movie(images, 1 / float(self.time_step), descriptions, name=name)


if __name__ == '__main__':
    mm = MovieMaker()
    movie = mm.render_movie()
    movie.save_as_avi_movie('movie_maker_test.avi')
    movie.save('movie_maker_test.pickle')
    movie2 = load_movie('movie_maker_test.pickle')
    movie2.save_as_avi_movie('movie_maker_test2.avi')



