"""Temporal signal generator"""

from math import pi
from numpy import arange
from numpy import array
from numpy import concatenate
from numpy import sin
from numpy import zeros
from random import Random

def make_sin(ampl=1, period=2 * pi, offset=0, mean=0):
    """Build a parameterised sinusoidal function"""
    def f(x):
        return ampl * sin((x - offset) * 2 * pi / period) + mean
    return f

def make_harmonic(parameters):
    """Build a parameterised sum of sinusoidal functions

    parameters are a sequence of sinusoidal parameters dict:
        - ampl: amplitude
        - period: period in time steps
        - offset: temporal offset in time steps
        - mean: mean value
    """
    def f(x):
        try:
            y = zeros(len(x))
        except TypeError:
            y = 0.0
        for param in parameters:
            if isinstance(param, dict):
                y += make_sin(**param)(x)
            else:
                y += make_sin(*param)(x)
        return y
    return f

def make_temporal_generator(funcs, max_length=None, dtype='f'):
    """Generate arrays of temporal data
    For each t from 0 to max_length generate arrays defined by::

        array([f(t) for f in funcs])

    If max_length is None generate arrays for ever.
    """
    t = 0
    while max_length is None or t < max_length:
        t += 1
        yield array([f(t) for f in funcs])


class OneDTemporalGenerator(object):
    """Utility component to generate 1D temporal signals

    Public attributes:

    :funcs:
        signal generating functions (empty tuple by default)

    :size:
        default size of a generated array of points

    :memory_size:
        default memory size for

    :differential:
        generate differential memory based training points

    :t0:
        time step number to use as temporal origine

    :dtype:
        default data type for generated numpy arrays

    :randomized:
        if True, generated training datasets are shuffle for better
        cross validation process
    """


    def __init__(self, funcs=(), size=100, memory_size=10, differential=False,
                 t0=0, dtype='f', randomized=True, seed=None):

        # internal random number generator instance
        self._rng = Random(seed)

        self.funcs = funcs
        self.size = size
        self.memory_size = memory_size
        self.differential = differential
        self.t0 = t0
        self.dtype = dtype
        self.randomized = randomized


    def seed(self, seed=None):
        """Seed the internal rng instance, use None for time-based seeding"""
        self._rng.seed(seed)

    def make_training_points(self, func, size=None, memory_size=None,
                             differential=None, t0=None):
        """Build a list of training points for a given function

        Each training point is a pair (y, x) where y is the next value and
        x is a array of length memory_size with previous values.
        """
        if size is None:
            size = self.size
        if memory_size is None:
            memory_size = self.memory_size
        if differential is None:
            differential = self.differential
        if t0 is None:
            t0 = self.t0

        time_range = arange(t0, t0 + memory_size + size)
        values = func(time_range)
        if differential:
            diff = array(values[1:]) - array(values[:-1])
            training_points = [
                (concatenate((diff[t:t+memory_size-1], values[t:t+memory_size])),
                 values[t+memory_size])
                for t in xrange(size)]

        else:
            training_points = [(values[t:t+memory_size], values[t+memory_size])
                               for t in xrange(size)]
        return training_points

    def make_training_set(self, funcs=None, randomized=None, size=None,
                          memory_size=None, differential=None, t0=None):
        """Build a training set out of list of a 1D-temporal functions

        Return x, y
        """
        if funcs is None:
            funcs = self.funcs
        if randomized is None:
            randomized = self.randomized

        training_points = []
        for f in funcs:
            training_points.extend(self.make_training_points(
                f, size, memory_size, differential, t0))

        if randomized:
            # shuffle again to mix points coming from different functions
            self._rng.shuffle(training_points)

        x = [x_t for x_t, _ in training_points]
        y = [y_t for _, y_t in training_points]

        return x, y


# Create one instance, seeded from current time, and export its methods
# as module-level functions.  The functions share state across all uses
# (both in the user's code and in the Python libraries), but that's fine
# for most programs and is easier for the casual user than making them
# instantiate their own Random() instance.

_inst = OneDTemporalGenerator()
seed = _inst.seed
make_training_points = _inst.make_training_points
make_training_set = _inst.make_training_set

