# (C) Copyright 2007 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""Utility to graph a gumby brain"""
import os
import shutil

from yapgvb import Graph
from tempfile import mkstemp, mkdtemp
from PIL import Image

class UidGen(object):
    """Gives unique ids to items"""

    def __init__(self):
        self._count = 0
        self._labels = dict()

    def __call__(self, item):
        label = self._labels.get(item)
        if label is None:
            self._count += 1
            self._labels[item] = str(self._count)
            return self._labels[item]
        else:
            return label


class BrainGrapher(object):
    """Utility class to graph a brain state"""

    avi_cmd_pattern = ("mencoder 'mf://%s/*.png' -mf type=png:fps=%0.2f "
                       "-ovc lavc -lavcopts vcodec=mpeg4 -oac copy -o %s; "
                       "ffmpeg2theora -p pro %s")

    gif_cmd_pattern = "convert -delay %d %s/*.png %s"

    def __init__(self, brain, fps=.5):
        self.brain = brain
        self.snapshots = []
        self.fps = fps

    def graph(self):
        """Build a graph out of the current brain state"""
        g = Graph()

        node_uid = UidGen()
        g_nodes = dict()

        # add the nodes
        for i, layer in enumerate(self.brain.get_ranked_layers()):
            sg = g.add_node(str(i))
            sg.label = u'Layer %d\ndimensions: %s' % \
                    (i, 'x'.join([str(d) for d in layer.dimensions]))
            sg.shape = 'rectangle'
            sg.width = 2

            # associate graph nodes to layers
            g_nodes[layer] = sg


        # add the edges
        for layer in self.brain:
            source = self.brain.get_source_layer(layer)
            if source is not None:
                g_nodes[layer] << g_nodes[source]

        g.fontsize = 3
        g.layout('dot')
        return g

    def snapshot(self):
        """Take a new snapshot of the current brain state"""
        self.snapshots.append(self.graph())

    def save_as_avi_movie(self, filename):
        directory = mkdtemp()
        for i, g in enumerate(self.snapshots):
            tmpfile = os.path.join(directory, 'snapshot_%04d.png' % i)
            g.render(tmpfile)
        os.system(self.avi_cmd_pattern % (
            directory, self.fps, filename, filename))
        shutil.rmtree(directory)

    def save_as_animated_gif(self, filename):
        directory = mkdtemp()
        for i, g in enumerate(self.snapshots):
            tmpfile = os.path.join(directory, 'snapshot_%04d.png' % i)
            g.render(tmpfile)
        os.system(self.gif_cmd_pattern % (
            1 / self.fps * 100, directory, filename))
        shutil.rmtree(directory)


def node_color(node):
    base_color = 'lightblue'
    if node.get_h_output() is None or node.get_t_output() is None:
        return base_color
    delta = abs(node.get_h_output() - node.get_t_output())
    if delta < 0.1:
        return base_color
    elif delta < 0.3:
        return base_color + '1'
    elif delta < 0.5:
        return base_color + '2'
    elif delta < 1.:
        return base_color + '3'
    else:
        return base_color + '4'

def node_label(node):
    label = "h %s\nt %s" % (
        ("%0.1f [%d] [%d]" % (node.get_h_output() or 0,
                              len(node.h_input_history),
                              len(node.h_output_history))).rjust(12),
        ("%0.1f [%d] [%d]" % (node.get_t_output() or 0,
                              len(node.t_input_history),
                              len(node.t_output_history))).rjust(12),
    )
    return label


def graph_brain(brain, filename):
    """Utility function to draw the graphical representation of a brain"""
    grapher = BrainGrapher(brain)
    grapher.graph().render(filename)


def show_brain(brain):
    """Display the graphed brain thanks to PIL.Image.show()"""
    _, filename = mkstemp('.png')
    graph_brain(brain, filename)
    Image.open(filename).show()
    os.remove(filename)

