# (C) Copyright 2008 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""Historical data management utilities"""

from collections import deque
from itertools import izip
from numpy import array
from numpy import zeros

class History(object):
    """Historical data management utility"""

    dimensions = None

    neighborhood = None

    size = 10

    def __init__(self, dimensions, neighborhood=None, size=None):
        assert len(dimensions) != 0, "at least one dimension required"
        for i, dim in enumerate(dimensions):
            assert dim > 0, "%dth dimension must be > 0" % i

        self.dimensions = dimensions
        if neighborhood is None:
            self.neighborhood = (3,) * len(dimensions)
        else:
            self.neighborhood = neighborhood

        if size is not None:
            self.size = size

        assert self.size > 0, "history size must be at least 1"

        # build a storage as a deque of arrays of dimensiens + zeros padded
        # extensions that will help extract neighborhoods of cells
        self._extended_dimensions = tuple(
            d + (n - 1) for d, n in izip(dimensions, self.neighborhood))
        self.reset()

        # pre-compute slice objects for later usage
        self._offset = offset = tuple((n - 1) / 2 for n in self.neighborhood)
        self._data_slice = [slice(o, d + o)
                            for o, d in izip(offset, dimensions)]

    def reset(self):
        """Zero all historical data"""
        self._history = deque(zeros(self._extended_dimensions)
                              for _ in xrange(self.size))

    def __len__(self):
        return self.size

    def __iter__(self):
        return (self[i] for i in xrange(self.size))

    def record(self, data):
        """Store new data at first position and forget the oldest record

        raise ValueError is data cannot be reshaped to the right dimensions
        """
        # check
        if (data.shape != self.dimensions):
            try:
                data = data.reshape(self.dimensions)
            except ValueError:
                raise ValueError("recorded data must accept shape %r" %
                                 (self.dimensions,))

        # pad data with zeros
        extended_data = zeros(self._extended_dimensions)
        extended_data[self._data_slice] = data

        # record new padded data and forget the past
        self._history.appendleft(extended_data)
        self._history.pop()

    def __getitem__(self, i):
        return self._history[i][self._data_slice]

    def get_neighborhood(self, center_index, time_index=0):
        """Extract"""
        if not all([ center_index[i] < self.dimensions[i]
                     for i,c in enumerate(center_index)]):
            raise ValueError("You requested a neighborhood center "
                             "outside the layer.")
        neighborhood_slice = [slice(c, c + n)
                             for n, c in izip(self.neighborhood, center_index)]
        return self._history[time_index][neighborhood_slice]


