# (C) Copyright 2007-2008 Olivier Grisel
# Author: Olivier Grisel <olivier.grisel@ensta.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id$
"""The SVM component: auto-parameterization, train and predict"""

from itertools import chain
import logging
from numpy import array
from numpy import zeros
from thread import get_ident

from svm.kernel import Linear
from svm.kernel import Polynomial
from svm.kernel import RBF
from svm.kernel import Sigmoid
from svm.regression import EpsilonRegressionModel
from svm.regression import NuRegressionModel
from svm import RegressionDataSet
from svm import TestDataSet

from processing import Pool


class Holder(object):
    """XXX: hack to workaround memory management bug in pyLibSvm"""

    def __init__(self):
        self._map = dict()

    def hold(self, dataset):
        self._map[get_ident()] = dataset


class SvmModel(object):
    """Base implementation of a SVM model

    >>> m = SvmModel()
    >>> m.kernel_type
    'rbf'
    >>> m.model_type
    'nu'

    >>> m.gamma = 10
    >>> m.input_size = 2

    >>> from numpy import array, abs

    >>> inputs = []
    >>> inputs.append(array([0., 0.]))
    >>> inputs.append(array([1., 0.]))
    >>> inputs.append(array([0., 1.]))
    >>> inputs.append(array([1., 1.]))
    >>> targets = [0., 1., 1., 0.]

    >>> m.train(inputs, targets)
    >>> predicted = m.predict(inputs)

    >>> abs(targets - predicted).sum()
    0.00...

    """
    logger = logging.getLogger("gumby.model.Model")

    #
    # Model parameter search space
    #

    model_types = (
        ('nu', NuRegressionModel),
        ('epsilon', EpsilonRegressionModel),
    )

    kernel_types = (
        ('linear', Linear),
        ('polynomial', Polynomial),
        ('rbf', RBF),
        ('sigmoidal', Sigmoid),
    )

    #
    # Hyper parameters default values and ranges
    #

    cost_range = (-1, 3)

    cost_base = 10.

    cost = 10

    gamma_range = (-3, 3)

    gamma_base = 2.

    gamma = 4.

    coef_range = (1, 3)

    coef_base = 2.

    coef = None

    degree_range = (2, 3)

    degree = None

    #
    # Model parameters default values
    #

    epsilon = 0.1

    nu = 0.05 # target 5% of the training points to be kept as SV

    #
    # Parallel computing
    #

    # number or processes in the pool
    worker_pool_size = 1 # TODO: reenable when pickling fixed

    # minimum input size per worker for parallel prediction
    parallel_predict_min = 400

    def _cost_range(self):
        return (self.cost_base ** n for n in xrange(*self.cost_range))

    def _gamma_range(self, kt_name, gamma_from_data, gamma):
        if kt_name == 'linear':
            return (None,)
        else:
            if gamma_from_data:
                return (gamma,)
            else:
                return (self.gamma_base ** n for n in xrange(*self.gamma_range))

    def _coef_range(self, kt_name):
        if kt_name in ('linear', 'rbf'):
            return (None,)
        else:
            return chain(
                (- self.coef_base ** n for n in xrange(*self.coef_range)),
                (  self.coef_base ** n for n in xrange(*self.coef_range)))

    def _degree_range(self, kt_name):
        if kt_name in ('linear', 'rbf', 'sigmoidal'):
            return (None,)
        else:
            return xrange(*self.degree_range)

    #
    # getters / setters for model and kernel
    #

    def _set_model_type(self, mt_name):
        mt_name = mt_name.lower()
        model_type = dict(self.model_types).get(mt_name)
        if model_type is not None:
            self._mt_name = mt_name
            self._mt = model_type
            self._predictor = None
        else:
            raise ValueError("invalid model type '%s'; must be one of %s" % (
                mt_name, [name for name, _ in self.model_types]))

    def _get_model_type(self):
        return self._mt_name

    model_type = property(_get_model_type, _set_model_type)

    def _set_kernel_type(self, kt_name):
        kt_name = kt_name.lower()
        kernel_type = dict(self.kernel_types).get(kt_name)
        if kernel_type is not None:
            self._kt_name = kt_name
            self._kt = kernel_type
            self._predictor = None
        else:
            raise ValueError("invalid kernel type '%s'; must be one of %s" % (
                kt_name, [name for name, _ in self.kernel_types]))

    def _get_kernel_type(self):
        return self._kt_name

    kernel_type = property(_get_kernel_type, _set_kernel_type)

    def __init__(self, input_size=10, model_type='nu', kernel_type='rbf',
                 worker_pool_size=None):

        self.input_size = input_size
        self.model_type = model_type
        self.kernel_type = kernel_type

        self._holder = Holder()
        self._predictor = None

        if worker_pool_size is not None:
            self.worker_pool_size = worker_pool_size
        self._worker_pool = Pool(self.worker_pool_size)

    def __repr__(self):
        attrs = ('input_size', 'model_type', 'kernel_type')
        return 'SvmModel(%s)' % ', '.join("%s=%r" % (attr, getattr(self, attr))
                                          for attr in attrs)

    def _check_sizes(self, inputs, outputs=None):
        if outputs is not None and len(inputs) != len(outputs):
            raise ValueError("inputs has of size %d and outputs has size %d" %
                            (len(inputs), len(outputs)))
        elif len(inputs) == 0:
            raise ValueError("inputs cannot be empty")

        input_size = len(inputs[0])
        if input_size != self.input_size:
            raise BadDimensionError("got input of size %d instead of %d" % (
                input_size, self.input_size))

    def is_trained(self):
        return self._predictor is not None

    def train(self, inputs, outputs, gamma_from_data=True):
        """Train the internal predictor to fit the given data"""

        self._check_sizes(inputs, outputs)

        dataset = self._make_training_set(inputs, outputs)
        if gamma_from_data:
            self.gamma = dataset.gamma
        model = self._make_model()

        self._predictor = model.fit(dataset)

        # perform model compaction if available
        try:
            self._predictor.compact()
        except NotImplementedError:
            # TODO: add info log here
            pass

    def _make_model(self):
        kargs = [a for a in (self.degree, self.gamma, self.coef)
                   if a is not None]
        kernel = self._kt(*kargs)

        model_kw = {'cost': self.cost}
        if self.model_type == 'nu':
            model_kw['nu'] = self.nu
        elif self.model_type == 'epsilon':
            model_kw['epsilon'] = self.epsilon
        return self._mt(kernel, **model_kw)

    def _make_training_set(self, inputs, outputs):
        # XXX: for some reason we get memory problems (even segfaults)
        # is the train data is not stored as some kind of instance attribute
        dataset = RegressionDataSet(outputs, inputs)
        self._holder.hold(dataset)
        return dataset

    def cross_validate(self, inputs, outputs, nr_folds=5):
        """Perform ``nr_folds`` cross validation w.r.t. to common parameters

        Return the mean squarred error and the squared correlation coefficient.
        """
        self._check_sizes(inputs, outputs)

        model = self._make_model()
        dataset = self._make_training_set(inputs, outputs)
        return model.cross_validate(dataset, nr_fold=nr_folds)

    def predict(self, inputs):
        """Predict the outcome of inputs

        Return an array of zeros if the model has been previously trained.
        """
        self._check_sizes(inputs)
        if self._predictor is None:
            return zeros(len(inputs))

        # TODO: fix me, I cannot get pickled!
        def predict_job(sub_inputs):
            testdata = TestDataSet(sub_inputs)
            return self._predictor.predict(testdata)

        chunk_size, remaining = divmod(len(inputs), self.worker_pool_size)
        if remaining:
            chunk_size += 1
        self.logger.debug("prediction chunk_size is %d", chunk_size)
        if chunk_size > self.parallel_predict_min and self.worker_pool_size > 1:
            # split the prediction taks into parallel processing sub tasks
            chunks = [inputs[i:i+chunk_size]
                      for i in xrange(self.worker_pool_size)]
            predictions = []
            for chunked_predictions in self._worker_pool.map(predict_job, chunks):
                predictions.extend(chunked_predictions)
            return array(predictions)

        else:
            # the overrhead is not worth the parallelization speed up
            testdata = TestDataSet(inputs)
            return array(self._predictor.predict(testdata))

    def find_parameters(self, inputs, outputs, do_model_type=False,
                        gamma_from_data=True, do_kernel_type=False, nr_folds=3):
        """Perform exponential grid optimization of model parameters"""

        self._check_sizes(inputs, outputs)

        if do_model_type:
            # optimize model type as well
            model_types = self.model_types
        else:
            # use current model type
            model_types = [(self._mt_name, self._mt)]

        if do_kernel_type:
            # optimize kernel type as well
            kernel_types = self.kernel_types
        else:
            # use current kernel type
            kernel_types = [(self._kt_name, self._kt)]

        dataset = self._make_training_set(inputs, outputs)

        # build the parameters grid generator
        params = ((mt, mt_name, kt, kt_name, cost, gamma, coef, degree)
                  for mt_name, mt in model_types
                  for kt_name, kt in kernel_types
                  for cost in self._cost_range()
                  for gamma in self._gamma_range(kt_name, gamma_from_data,
                                                 dataset.gamma)
                  for coef in self._coef_range(kt_name)
                  for degree in self._degree_range(kt_name))

        # perform the grid search
        best_mse, best_sc = None, None

        for mt, mt_name, kt, kt_name, cost, gamma, coef, degree in params:
            kargs = [a for a in (degree, gamma, coef) if a is not None]
            kernel = kt(*kargs)
            model = mt(kernel, cost=cost)
            mse, sc = model.cross_validate(dataset, nr_folds)

            if best_mse is None or best_mse > mse:
                # these parameters are better than ant previously found
                best_mse, best_sc = mse, sc
                mt_name_opt, kt_name_opt, cost_opt = mt_name, kt_name, cost
                gamma_opt, coef_opt, degree_opt = gamma, coef, degree

        self.logger.info("Optimal parameters found: mt_name=%s, "
                         "kt_name=%s, cost=%s, gamma=%s, coef=%s, degree=%s",
                         mt_name_opt, kt_name_opt, cost_opt, gamma_opt,
                         coef_opt, degree_opt)

        # save best parameters found
        self.model_type = mt_name_opt
        self.kernel_type = kt_name_opt
        self.cost = cost_opt
        self.gamma = gamma_opt
        self.coef = coef_opt
        self.degree = degree_opt

        # return training statistics
        return best_mse, best_sc


class BadDimensionError(Exception):
    pass

class UntrainedModelError(Exception):
    pass
