import unittest

from gumby.brain import Brain
from gumby.graphviz import show_brain
from numpy import random

class BrainTestCase(unittest.TestCase):

    def _make_brain(self):
        """Build a sample test brain"""
        # empty brain
        brain = Brain()

        # add 3 input nodes sharing the same t_model
        n1 = brain.add_node()
        brain.add_node(share_models=n1)
        brain.add_node(share_models=n1)
        group1 = brain.get_h_group(n1)

        # add a second layer
        n2 = brain.add_nodes(3)[0]
        group2 = brain.get_h_group(n2)

        # connect the second layer to the first with at most 2 links per target
        # node
        brain.connect_groups(group1, group2, max=2)

        # add a third layer
        n3 = brain.add_nodes(3)[0]
        group3 = brain.get_h_group(n3)

        # fully connect the third level to the second level
        brain.connect_groups(group2, group3)

        self.brain = brain
        self.n1 = n1
        self.n2 = n2
        self.n3 = n3
        self.group1 = group1
        self.group2 = group2
        self.group3 = group3

    def setUp(self):
        # seed the numpy rng
        random.seed(10)

        # build the brain itsel
        self._make_brain()

        # initialize the brain
        self.brain.initialize()

    def _show(self, brain=None):
        """Debug helper"""
        brain = brain or self.brain
        show_brain(brain)

    def test_default_structure(self):
        self.assertEquals(9, len(self.brain))
        self.assertEquals(3, len(self.group1))
        self.assertEquals(3, len(self.group2))
        self.assertEquals(3, len(self.group3))

        self.assertEquals([], self.brain.get_source_nodes(self.n1))
        self.assertEquals(2, len(self.brain.get_target_nodes(self.n1)))

        self.assertEquals(2, len(self.brain.get_source_nodes(self.n2)))
        self.assertEquals(3, len(self.brain.get_target_nodes(self.n2)))

        self.assertEquals(3, len(self.brain.get_source_nodes(self.n3)))
        self.assertEquals(0, len(self.brain.get_target_nodes(self.n3)))

    def test_default_node_histories(self):
        for node in self.brain:
            self.assertEquals([], node.h_input_history)
            self.assertEquals([], node.h_output_history)

            self.assertEquals([], node.t_input_history)
            self.assertEquals([], node.t_output_history)

    def test_default_h_groups(self):
        h_groups = [tuple(nodes)
                    for _, nodes in self.brain.get_ranked_h_groups()]
        self.assertEquals(self.group1, h_groups[0])
        self.assertEquals(self.group2, h_groups[1])
        self.assertEquals(self.group3, h_groups[2])

    def test_default_t_groups(self):
        t_groups = [tuple(nodes)
                    for _, nodes in self.brain.get_ranked_t_groups()]
        self.assertEquals(self.group1, t_groups[2])
        self.assertEquals(self.group2, t_groups[1])
        self.assertEquals(self.group3, t_groups[0])

    def test__md_indexes_to_od_index(self):
        b = Brain()

        res = b._md_indexes_to_od_index((0, 0), (3, 5))
        self.assertEquals(0, res)

        res = b._md_indexes_to_od_index((0, 1), (3, 5))
        self.assertEquals(1, res)

        res = b._md_indexes_to_od_index((1, 0), (3, 5))
        self.assertEquals(5, res)

        res = b._md_indexes_to_od_index((1, 2), (3, 5))
        self.assertEquals(7, res)

        res = b._md_indexes_to_od_index((4, 3, 20), (10, 20, 30))
        self.assertEquals( 4 * 20 * 30 + 3 * 30 + 20, res)

    def test__rec_build_md_indexes(self):
        b = Brain()

        indexes = b._rec_build_md_indexes([], [], [xrange(0, 2), xrange(0, 3)])
        self.assertEquals([0, 0], indexes.pop(0))
        self.assertEquals([0, 1], indexes.pop(0))
        self.assertEquals([0, 2], indexes.pop(0))
        self.assertEquals([1, 0], indexes.pop(0))
        self.assertEquals([1, 1], indexes.pop(0))
        self.assertEquals([1, 2], indexes.pop(0))

    def test__get_submap_indexes(self):
        b = Brain()
        indexes = b._get_submap_indexes((0, 0), (2, 2), (2, 2))
        self.assertEquals([0, 1, 2, 3], indexes)

        indexes = b._get_submap_indexes((0, 0), (3, 3), (2, 2))
        self.assertEquals([0, 1, 3, 4], indexes)

        indexes = b._get_submap_indexes((0, 1), (3, 3), (2, 2))
        self.assertEquals([ 1, 2, 4, 5], indexes)

        indexes = b._get_submap_indexes((1, 1), (3, 3), (2, 2))
        self.assertEquals([ 4, 5, 7, 8], indexes)

    def test__get_map_indexes(self):
        b = Brain()
        indexes = list(b._get_map_indexes((2, 2), (2, 2), (0, 0)))
        self.assertEquals(1, len(indexes))
        self.assertEquals([0, 1, 2, 3], list(indexes[0]))

        indexes = list(b._get_map_indexes((2, 3), (2, 2), (0, 0)))
        self.assertEquals(2, len(indexes))
        self.assertEquals([0, 1, 3, 4], list(indexes[0]))
        self.assertEquals([1, 2, 4, 5], list(indexes[1]))

        indexes = list(self.brain._get_map_indexes((3, 5), (2, 3), (0, 1)))
        self.assertEquals(4, len(indexes))
        self.assertEquals([0, 1, 2, 5, 6, 7], list(indexes[0]))
        self.assertEquals([2, 3, 4, 7, 8, 9], list(indexes[1]))
        self.assertEquals([5, 6, 7, 10, 11, 12], list(indexes[2]))
        self.assertEquals([7, 8, 9, 12, 13, 14], list(indexes[3]))

        indexes = list(self.brain._get_map_indexes((1, 1, 1), (1, 1, 1),
                                                   (0, 0, 0)))
        self.assertEquals(1, len(indexes))
        self.assertEquals([0], list(indexes[0]))

        indexes = list(self.brain._get_map_indexes((1, 2, 1), (1, 2, 1),
                                                   (0, 0, 0)))
        self.assertEquals(1, len(indexes))
        self.assertEquals([0, 1], list(indexes[0]))

        indexes = list(self.brain._get_map_indexes((1, 3, 1), (1, 2, 1),
                                                   (0, 0, 0)))
        self.assertEquals(2, len(indexes))
        self.assertEquals([0, 1], list(indexes[0]))
        self.assertEquals([1, 2], list(indexes[1]))

    def test_add_map(self):
        brain = Brain()
        source_nodes = brain.add_nodes(n=100)
        middle_nodes = brain.add_map(source_nodes, (10, 10), (5, 5), (1, 1))
        self.assertEquals(9, len(middle_nodes))

        for mn in middle_nodes:
            self.assertEquals(25, len(brain.get_source_nodes(mn)))

        top_nodes = brain.add_map(middle_nodes, (3, 3), (2, 2))
        self.assertEquals(4, len(top_nodes))

        for tn in top_nodes:
            self.assertEquals(4, len(brain.get_source_nodes(tn)))

    def test_add_map2(self):
        brain = Brain()
        source_nodes = brain.add_nodes(n=200)
        middle_nodes = brain.add_map(source_nodes, (5, 5, 4), (2, 2, 4),
                                     (1, 1, 0))
        self.assertEquals(4, len(middle_nodes))

        for mn in middle_nodes:
            self.assertEquals(16, len(brain.get_source_nodes(mn)))

        top_nodes = brain.add_map(middle_nodes, (2, 2), (2, 2))
        self.assertEquals(1, len(top_nodes))

        for tn in top_nodes:
            self.assertEquals(4, len(brain.get_source_nodes(tn)))

    def test_set_input(self):
        # input nodes are node without other nodes as sources
        self.assertEquals(self.group1, tuple(self.brain.input_nodes))

        # test wrong input size
        self.assertRaises(ValueError, self.brain.set_input, [1., 1.])

        # test input setting when input size is correct
        input_values = [.2, .9, -.8]
        self.brain.set_input(input_values)

        # measure node values
        group1_h_outputs = [n.get_h_output() for n in self.group1]
        self.assertEquals(input_values, group1_h_outputs)

        # other nodes should remain undefined
        group2_h_outputs = [n.get_h_output() for n in self.group2]
        self.assertEquals([None, None, None], group2_h_outputs)

        group3_h_outputs = [n.get_h_output() for n in self.group2]
        self.assertEquals([None, None, None], group3_h_outputs)

        # t values should remain unaffected
        for node in self.brain:
            self.assertEquals(None, node.get_t_output())

        # all histories should be empty exept the h_output values of the input
        # nodes
        group1_h_output_histories = [n.h_output_history for n in self.group1]
        self.assertEquals([[.2], [.9], [-.8]], group1_h_output_histories)

        for node in self.brain:
            self.assertEquals([], node.h_input_history)
            if node not in self.group1:
                self.assertEquals([], node.h_output_history)
            self.assertEquals([], node.t_input_history)
            self.assertEquals([], node.t_output_history)

    def test_h_inference(self):
        # set inputs
        input_values = [.2, .9, -.8]
        self.brain.set_input(input_values)

        # trigger the h inference step
        self.brain.perform_h_inference()

        for node in self.brain:
            # at this point all h values should be defined while t values
            # should still be undefined:
            self.assertNotEquals(None, node.get_h_output())
            self.assertEquals(None, node.get_t_output())

            # similarly all h output histories should a have one value:
            self.assertEquals([node.get_h_output()], node.h_output_history)

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(1, len(node.h_input_history))

            # all t histories should still be empty
            self.assertEquals(0, len(node.t_input_history))
            self.assertEquals(0, len(node.t_output_history))

    def test_t_inference(self):
        # set inputs
        input_values = [.2, .9, -.8]
        self.brain.set_input(input_values)

        # t-inference cannot be performed unless h-inference has been performed
        # first
        self.brain.perform_h_inference()

        # trigger the t inference step
        self.brain.perform_t_inference()

        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # similarly all h output histories should a have one value:
            self.assertEquals([node.get_h_output()], node.h_output_history)

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(1, len(node.h_input_history))

            # all t-histories should now have on element
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(1, len(node.t_output_history))

        # now all inference steps are done we have a prediction
        self.assertEquals(3, len(self.brain.get_latest_output()))

    def test_train(self):

        # let us first build some history with the get prediction method
        inputs = [[.2, .9, -.8], [-.7, -.1, .2], [-.3, -.1, -.5]]
        for input in inputs:
            prediction = self.brain.get_prediction(input)
            self.assertEquals(3, len(prediction))

        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())


            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            self.assertEquals(3, len(node.h_output_history))
            self.assertEquals(3, len(node.t_input_history))
            self.assertEquals(3, len(node.t_output_history))


        # build a generator that trains the models iteratively
        trainer = self.brain.get_training_generator()

        # train the first t-model, ie the one of group3
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            if node in self.group3:
                # inference history has been consumed by the trainer except
                # one t input value that will be used by the trainer the next
                # time (matching h output has not yet been observed)
                self.assertEquals(1, len(node.t_input_history))
                self.assertEquals(0, len(node.h_output_history))
            else:
                self.assertEquals(3, len(node.h_output_history))
                self.assertEquals(3, len(node.t_input_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the second t-model, ie the one of group2
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            if node in self.group3 or node in self.group2:
                # inference history has been consumed by the trainer except
                # one t input value that will be used by the trainer the next
                # time (matching h output has not yet been observed)
                self.assertEquals(1, len(node.t_input_history))
                self.assertEquals(0, len(node.h_output_history))
            else:
                self.assertEquals(3, len(node.h_output_history))
                self.assertEquals(3, len(node.t_input_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the third t-model, ie the one of group1
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the first h-model, ie the one of group1 (does nothing apart
        # from cleaning the useless t-output history content)
        trainer.next()

        # train the second h-model, ie the one of group2
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1 or node in self.group2:
                self.assertEquals(0, len(node.h_input_history))
                self.assertEquals(0, len(node.t_output_history))
            else:
                self.assertEquals(3, len(node.h_input_history))
                self.assertEquals(3, len(node.t_output_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))

        # train the third h-model, ie the one of group3
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # all h-model related inference history is consumed
            self.assertEquals(0, len(node.h_input_history))
            self.assertEquals(0, len(node.t_output_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))

        # now we have done a complete training cycle let us do some more
        # inference to refill the histories an further use them as training data
        for input in inputs:
            prediction = self.brain.get_prediction(input)
            self.assertEquals(3, len(prediction))

        # let us now retrain, same as a previously except the t-input histories
        # have one element that has been kept
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())


            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            self.assertEquals(3, len(node.h_output_history))
            self.assertEquals(4, len(node.t_input_history))
            self.assertEquals(3, len(node.t_output_history))


        # build a generator that trains the models iteratively
        trainer = self.brain.get_training_generator()

        # train the first t-model, ie the one of group3
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            if node in self.group3:
                # inference history has been consumed by the trainer except
                # one t input value that will be used by the trainer the next
                # time (matching h output has not yet been observed)
                self.assertEquals(1, len(node.t_input_history))
                self.assertEquals(0, len(node.h_output_history))
            else:
                self.assertEquals(3, len(node.h_output_history))
                self.assertEquals(4, len(node.t_input_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the second t-model, ie the one of group2
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            if node in self.group3 or node in self.group2:
                # inference history has been consumed by the trainer except
                # one t input value that will be used by the trainer the next
                # time (matching h output has not yet been observed)
                self.assertEquals(1, len(node.t_input_history))
                self.assertEquals(0, len(node.h_output_history))
            else:
                self.assertEquals(3, len(node.h_output_history))
                self.assertEquals(4, len(node.t_input_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the third t-model, ie the one of group1
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1:
                self.assertEquals(0, len(node.h_input_history))
            else:
                self.assertEquals(3, len(node.h_input_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))

            self.assertEquals(3, len(node.t_output_history))

        # train the first h-model, ie the one of group1 (does nothing apart
        # from cleaning the useless t-output history content)
        trainer.next()

        # train the second h-model, ie the one of group2
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # except for the input nodes, the h histories should hold
            # the following expected values
            if node in self.group1 or node in self.group2:
                self.assertEquals(0, len(node.h_input_history))
                self.assertEquals(0, len(node.t_output_history))
            else:
                self.assertEquals(3, len(node.h_input_history))
                self.assertEquals(3, len(node.t_output_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))

        # train the third h-model, ie the one of group3
        trainer.next()
        for node in self.brain:
            # at this point all h and t values should be defined
            self.assertNotEquals(None, node.get_h_output())
            self.assertNotEquals(None, node.get_t_output())

            # all h-model related inference history is consumed
            self.assertEquals(0, len(node.h_input_history))
            self.assertEquals(0, len(node.t_output_history))

            # inference history has been consumed by the trainer except
            # one t input value that will be used by the trainer the next
            # time (matching h output has not yet been observed)
            self.assertEquals(1, len(node.t_input_history))
            self.assertEquals(0, len(node.h_output_history))



def test_suite():
    suite = unittest.TestSuite()
    # disabled while not yet refactored
    #suite.addTests(unittest.makeSuite(BrainTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')

