import unittest

import random
from numpy import zeros
from numpy import array
from gumby.brain import Layer
from gumby.datagenerator.temporal import make_harmonic
from gumby.datagenerator.temporal import make_temporal_generator

# for debugging only
from pylab import concatenate
from pylab import figure
from pylab import grid
from pylab import plot
from pylab import show
from pylab import subplot
from pylab import title
from pylab import xlabel
from pylab import ylabel

class HarmonicSignalLayerTestCase(unittest.TestCase):

    # parameters are by order:
    #    - amplitude
    #    - period in time steps
    #    - offset in time steps
    #    - mean value

    params = [
        [
            (1, 50, 0, -0.3)
        ],
        [
            (3, 100, 0, 0.2),
            (0.5, 50, 20),
        ],
        [
            (0.5, 203, 0),
            (0.5, 200, 20),
            (0.5, 25, 5),
        ],
        [
            (0.5, 50, 0, 1),
            (0.5, 20, 12),
            (0.2, 30, 56),
            (0.2, 35, 56),
            (0.2, 36, 56),
            (0.2, 37, 52),
            #(0.2, 47, 52),
        ],
    ]

    def setUp(self):
        random.seed(0)
        self.layer = Layer(dimensions=(len(self.params),), history_size=300,
                           spatial_neighborhood=(1,),
                           temporal_neighborhood=100)

        funcs = [make_harmonic(param) for param in self.params]
        self.generator = make_temporal_generator(funcs)

    def test_untrained_prediction(self):
        expected = zeros(self.layer.dimensions)
        for t in xrange(self.layer.history_size * 2):
            pred = self.layer.predict(self.generator.next())
            self.assert_((expected == pred).all())

    def test_prediction(self):
        # load the layer with historical data
        self.test_untrained_prediction()
        err_before_training = self.layer.get_squared_prediction_error()

        # train it
        self.layer.model.cost = 10
        self.layer.train(max_sample_size=300)

        # run predictions on the trained layer
        predictions = [self.layer.predict(self.generator.next())]
        expected = [self.generator.next()]
        surprise = [self.layer.get_surprise(0)]
        for t in xrange(self.layer.history_size):
            predictions.append(self.layer.predict(expected[-1]))
            expected.append(self.generator.next())
            surprise.append(self.layer.get_surprise(0))

        err_after_training = self.layer.get_squared_prediction_error()

        # check that the error after training is smaller than before training
        self.assert_(err_after_training < err_before_training)

        # trying recursive predictions
        for t in xrange(self.layer.history_size * 2):
            predictions.append(self.layer.predict(predictions[-1]))
            expected.append(self.generator.next())
            surprise.append(self.layer.get_surprise(0))

        #self._plot(predictions, expected, surprise)

    def _plot(self, predictions, expected, surprise):
        predictions = array(predictions)
        expected = array(expected)

        #subplot(211)
        for i in xrange(len(self.params)):
            figure(i)
            grid(True)
            xlabel('time')
            ylabel('signal value')
            #subplot(211)
            plot(expected[:,i], 'r-')
            plot(predictions[:,i], 'b-')
            #plot(surprise, 'g-')

        #window_size = memory_size * 3
        #averaged_error = compute_error(predicted_future, expected_future,
        #                               window_size)
        #subplot(212)
        #xlabel('time')
        #ylabel('averaged error')
        #grid(True)
        #plot(averaged_error, 'b-')
        show()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests(unittest.makeSuite(HarmonicSignalLayerTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')

